import json
import platform
import re
from unittest import TestCase
from snapshottest import TestCase as SnapshotTestCase
from ain.ain import Ain
from ain.ain import Wallet
from ain.provider import JSON_RPC_ENDPOINT
from ain.utils import *
from ain.utils.v3keystore import *
from ain.errors import *
from ain.types import *
from .data import (
    mnemonic,
    message,
    v3KeystorePassword,
    v3KeystoreJSONList,
    tx,
    testNode,
    accountAddress,
    accountSk,
    transferAddress,
)
from .util import (
    asyncTest,
    getRequest,
    postRequest,
    waitUntilTxFinalized,
)

TX_PATTERN = re.compile("0x[0-9a-fA-F]{64}")
PY_VERSION = platform.python_version().replace(".", "")
APP_NAME = "bfan"

class TestNetwork(TestCase):
    def testChainId(self):
        ain = Ain(testNode)
        self.assertEqual(ain.chainId, 0)
        self.assertEqual(ain.wallet.chainId, 0)
        ain.setProvider(testNode, 2)
        self.assertEqual(ain.chainId, 2)
        self.assertEqual(ain.wallet.chainId, 2)

    def testSanitizeProviderURL(self):
        ain = Ain(testNode)
        with self.assertRaises(ValueError):
            ain.setProvider("")
        with self.assertRaises(ValueError):
            ain.setProvider("localhost:3000")
        noTrailingSlash = "http://localhost:3000"
        ain.setProvider(noTrailingSlash)
        self.assertEqual(ain.provider.apiEndPoint, noTrailingSlash + "/" + JSON_RPC_ENDPOINT)
        ain.setProvider(noTrailingSlash + "/")
        self.assertEqual(ain.provider.apiEndPoint, noTrailingSlash + "/" + JSON_RPC_ENDPOINT)

    @asyncTest
    async def testGetProtocolVersion(self):
        ain = Ain(testNode)
        self.assertNotEqual(await ain.net.getProtocolVersion(), None)
    
    @asyncTest
    async def testCheckProtocolVersion(self):
        ain = Ain(testNode)
        res = await ain.net.checkProtocolVersion()
        self.assertEqual(res["code"], 0)
        self.assertEqual(res["result"], True)

class TestWallet(TestCase):
    def testCreateAccount(self):
        ain = Ain(testNode)
        ain.wallet.create(2)
        self.assertEqual(ain.wallet.length, 2)

    def testAddFromV3Keystore(self):
        ain = Ain(testNode)
        keystoreAddress = ain.wallet.addFromV3Keystore(v3KeystoreJSONList[0], v3KeystorePassword)
        self.assertEqual(ain.wallet.length, 1)
        convertedV3Keystore = ain.wallet.accountToV3Keystore(keystoreAddress, v3KeystorePassword)
        derivedAddress = privateToAddress(convertedV3Keystore.toPrivateKey(v3KeystorePassword))
        self.assertEqual(derivedAddress, keystoreAddress)

    def testAddAccount(self):
        ain = Ain(testNode)
        ain.wallet.add(accountSk)
        self.assertEqual(ain.wallet.length, 1)
        with self.assertRaises(ValueError):
            ain.wallet.add("")
        self.assertEqual(ain.wallet.length, 1)

    def testAddAndSetDefaultAccount(self):
        ain = Ain(testNode)
        ain.wallet.addAndSetDefaultAccount(accountSk)
        self.assertEqual(ain.wallet.defaultAccount.private_key, accountSk)

    def testAddFromHDWallet(self):
        ain = Ain(testNode)
        ain.wallet.addFromHDWallet(mnemonic)
        self.assertEqual(ain.wallet.length, 1)

    def testRemoveAccount(self):
        ain = Ain(testNode)
        ain.wallet.add(accountSk)
        self.assertEqual(ain.wallet.length, 1)
        ain.wallet.remove(accountAddress)
        self.assertEqual(ain.wallet.length, 0)

    def testClearAccount(self):
        ain = Ain(testNode)
        ain.wallet.add(accountSk)
        ain.wallet.addFromHDWallet(mnemonic)
        self.assertEqual(ain.wallet.length, 2)
        ain.wallet.clear()
        self.assertEqual(ain.wallet.length, 0)

    def testSetDefaultAccount(self):
        ain = Ain(testNode)
        with self.assertRaises(ValueError):
            ain.wallet.setDefaultAccount(accountAddress)
        ain.wallet.add(accountSk)
        ain.wallet.setDefaultAccount(accountAddress.lower())
        self.assertEqual(ain.wallet.defaultAccount.address, accountAddress)
        ain.wallet.setDefaultAccount(accountAddress)
        self.assertEqual(ain.wallet.defaultAccount.address, accountAddress)
    
    def testRemoveDefaultAccount(self):
        ain = Ain(testNode)
        ain.wallet.addAndSetDefaultAccount(accountSk)
        self.assertEqual(ain.wallet.defaultAccount.private_key, accountSk)
        ain.wallet.removeDefaultAccount()
        self.assertEqual(ain.wallet.defaultAccount, None)

    def testSign(self):
        ain = Ain(testNode)
        with self.assertRaises(ValueError):
            ain.wallet.sign(message)
        ain.wallet.addAndSetDefaultAccount(accountSk)
        sig = ain.wallet.sign(message)
        self.assertEqual(ecVerifySig(message, sig, accountAddress), True)

    def testSignTransaction(self):
        ain = Ain(testNode)
        ain.wallet.addAndSetDefaultAccount(accountSk)
        sig = ain.wallet.signTransaction(tx)
        self.assertEqual(ain.wallet.verifySignature(tx, sig, accountAddress), True)
        self.assertEqual(ain.wallet.recover(sig), accountAddress)

    @asyncTest
    async def testGetBalance(self):
        ain = Ain(testNode)
        ain.wallet.addAndSetDefaultAccount(accountSk)
        self.assertGreaterEqual(await ain.wallet.getBalance(), 0)

    @asyncTest
    async def testTransferIsDryrunTrue(self):
        ain = Ain(testNode)
        ain.wallet.addAndSetDefaultAccount(accountSk)
        balanceBefore = await ain.wallet.getBalance()
        await ain.wallet.transfer(transferAddress, 100, nonce=-1, isDryrun=True)  # isDryrun = True
        balanceAfter = await ain.wallet.getBalance()
        self.assertEqual(balanceBefore, balanceAfter)  # NOT changed!

    @asyncTest
    async def testTransfer(self):
        ain = Ain(testNode)
        ain.wallet.addAndSetDefaultAccount(accountSk)
        balanceBefore = await ain.wallet.getBalance()
        await ain.wallet.transfer(transferAddress, 100, nonce=-1)
        balanceAfter = await ain.wallet.getBalance()
        self.assertEqual(balanceBefore - 100, balanceAfter)

    @asyncTest
    async def testTransferWithAZeroValue(self):
        ain = Ain(testNode)
        ain.wallet.addAndSetDefaultAccount(accountSk)
        balanceBefore = await ain.wallet.getBalance()
        try:
            await ain.wallet.transfer(transferAddress, 0, nonce=-1)  # zero value
            self.fail('should not happen')
        except ValueError as e:
            self.assertEqual(str(e), 'Non-positive transfer value.')
        balanceAfter = await ain.wallet.getBalance()
        self.assertEqual(balanceBefore, balanceAfter)  # NOT changed!

    @asyncTest
    async def testTransferWithANegativeValue(self):
        ain = Ain(testNode)
        ain.wallet.addAndSetDefaultAccount(accountSk)
        balanceBefore = await ain.wallet.getBalance()
        try:
            await ain.wallet.transfer(transferAddress, -0.1, nonce=-1)  # negative value
            self.fail('should not happen')
        except ValueError as e:
            self.assertEqual(str(e), 'Non-positive transfer value.')
        balanceAfter = await ain.wallet.getBalance()
        self.assertEqual(balanceBefore, balanceAfter)  # NOT changed!

    @asyncTest
    async def testTransferWithAValueOfUpTo6Decimals(self):
        ain = Ain(testNode)
        ain.wallet.addAndSetDefaultAccount(accountSk)
        balanceBefore = await ain.wallet.getBalance()
        await ain.wallet.transfer(transferAddress, 0.000001, nonce=-1)  # of 6 decimals
        balanceAfter = await ain.wallet.getBalance()
        self.assertEqual(balanceBefore - 0.000001, balanceAfter)

    @asyncTest
    async def testTransferWithAValueOfMoreThan6Decimals(self):
        ain = Ain(testNode)
        ain.wallet.addAndSetDefaultAccount(accountSk)
        balanceBefore = await ain.wallet.getBalance()
        try:
            await ain.wallet.transfer(transferAddress, 0.0000001, nonce=-1)  # of 7 decimals
            self.fail('should not happen')
        except ValueError as e:
            self.assertEqual(str(e), 'Transfer value of more than 6 decimals.')
        balanceAfter = await ain.wallet.getBalance()
        self.assertEqual(balanceBefore, balanceAfter)  # NOT changed!

    def testChainId(self):
        ain = Ain(testNode, 0)
        ain.wallet.addAndSetDefaultAccount(accountSk)

        sig = ain.wallet.signTransaction(tx)
        self.assertEqual(ain.wallet.verifySignature(tx, sig, accountAddress), True)
        self.assertEqual(ecVerifySig(tx, sig, accountAddress, 2), False)
        self.assertEqual(ain.wallet.recover(sig), accountAddress)

        ain.setProvider(testNode, 2)
        sig = ain.wallet.signTransaction(tx)
        self.assertEqual(ain.wallet.verifySignature(tx, sig, accountAddress), True)
        self.assertEqual(ecVerifySig(tx, sig, accountAddress, 0), False)
        self.assertEqual(ain.wallet.recover(sig), accountAddress)
            
class TestCore(TestCase):
    ain: Ain
    defaultAddr: str
    addr1: str
    addr2: str
    
    @classmethod
    @asyncTest
    async def setUpClass(cls):
        cls.ain = Ain(testNode, 0)
        cls.ain.wallet.addAndSetDefaultAccount(accountSk)
        newAccounts = cls.ain.wallet.create(2)
        cls.defaultAddr = cls.ain.wallet.defaultAccount.address
        cls.addr1 = newAccounts[0]
        cls.addr2 = newAccounts[1]
        getAddr = await getRequest(f"{testNode}/get_address")
        nodeAddr = getAddr["result"]
        stakeForApps = await postRequest(f"{testNode}/set", {
            "op_list": [
                {
                    "type": "SET_VALUE",
                    "ref": f"/staking/test{PY_VERSION}/{nodeAddr}/0/stake/{getTimestamp()}/value",
                    "value": 1
                },
                {
                    "type": "SET_VALUE",
                    "ref": f"/staking/{APP_NAME}{PY_VERSION}/{nodeAddr}/0/stake/{getTimestamp()}/value",
                    "value": 1
                }
            ],
            "nonce": -1
        })
        await waitUntilTxFinalized(testNode, stakeForApps["result"]["tx_hash"])

        createApps = await postRequest(f"{testNode}/set", {
            "op_list": [
                {
                    "type": "SET_VALUE",
                    "ref": f"/manage_app/test{PY_VERSION}/create/{getTimestamp()}",
                    "value": { "admin": { cls.defaultAddr: True } }
                },
                {
                    "type": "SET_VALUE",
                    "ref": f"/manage_app/{APP_NAME}{PY_VERSION}/create/{getTimestamp()}",
                    "value": { "admin": { cls.defaultAddr: True } }
                }
            ],
            "nonce": -1
        })
        await waitUntilTxFinalized(testNode, createApps["result"]["tx_hash"])

    @asyncTest
    async def test00GetBlock(self):
        block = await self.ain.getBlock(3)
        hash = block.get("hash", "")
        self.assertDictEqual(await self.ain.getBlock(hash), block)

    @asyncTest
    async def test00GetProposer(self):
        proposer = await self.ain.getProposer(1)
        block = await self.ain.getBlock(1)
        hash = block.get("hash", "")
        self.assertEqual(await self.ain.getProposer(hash), proposer)

    @asyncTest
    async def test00GetValidators(self):
        validators = await self.ain.getValidators(4)
        block = await self.ain.getBlock(4)
        hash = block.get("hash", "")
        self.assertDictEqual(await self.ain.getValidators(hash), validators)

    @asyncTest
    async def test00ValidateAppNameTrue(self):
        res = await self.ain.validateAppName("test_new")
        self.assertEqual(res["is_valid"], True)
        self.assertEqual(res["result"], True)
        self.assertEqual(res["code"], 0)
        self.assertTrue("protoVer" in res)

    @asyncTest
    async def test00ValidateAppNameFalse(self):
        raised = False
        try:
            await self.ain.validateAppName("app/path")
        except BlockchainError as e:
            self.assertEqual(e.code, 30601)
            self.assertEqual(e.message, "Invalid app name for state label: app/path")
            raised = True
        self.assertTrue(raised)

    @asyncTest
    async def test00SendTransactionIsDryrunTrue(self):
        op = SetOperation(
            type="SET_OWNER",
            ref=f"/apps/test{PY_VERSION}",
            value={
                ".owner": {
                    "owners": {
                        "*": {
                            "write_owner": True,
                            "write_rule": True,
                            "write_function": True,
                            "branch_owner": True
                        }
                    }
                }
            }
        )
        res = await self.ain.sendTransaction(TransactionInput(operation=op), True)  # isDryrun = True
        self.assertEqual(res["result"]["code"], 0)
        targetTxHash = res["tx_hash"]
        self.assertTrue(TX_PATTERN.fullmatch(targetTxHash) is not None)

        tx = await self.ain.getTransaction(targetTxHash)
        self.assertIsNone(tx)  # should be None

    @asyncTest
    async def test00SendTransaction(self):
        op = SetOperation(
            type="SET_OWNER",
            ref=f"/apps/test{PY_VERSION}",
            value={
                ".owner": {
                    "owners": {
                        "*": {
                            "write_owner": True,
                            "write_rule": True,
                            "write_function": True,
                            "branch_owner": True
                        }
                    }
                }
            }
        )
        res = await self.ain.sendTransaction(TransactionInput(operation=op))
        self.assertEqual(res["result"]["code"], 0)
        targetTxHash = res["tx_hash"]
        self.assertTrue(TX_PATTERN.fullmatch(targetTxHash) is not None)

        tx = await self.ain.getTransaction(targetTxHash)
        self.assertDictEqual(tx["transaction"]["tx_body"]["operation"], op.__dict__)

    @asyncTest
    async def test00SendSignedTransactionIsDryrunTrue(self):
        tx = TransactionBody(
            nonce=-1,
            gas_price=500,
            timestamp=getTimestamp(),
            operation=SetOperation(
                type="SET_OWNER",
                ref=f"/apps/{APP_NAME}{PY_VERSION}",
                value={
                    ".owner": {
                        "owners": {
                            "*": {
                                "write_owner": True,
                                "write_rule": True,
                                "write_function": True,
                                "branch_owner": True
                            }
                        }
                    }
                }
            )
        )

        sig = self.ain.wallet.signTransaction(tx)
        res = await self.ain.sendSignedTransaction(sig, tx, True)  # isDryrun = True
        targetTxHash = res["tx_hash"]
        self.assertFalse("code" in res)
        self.assertTrue(TX_PATTERN.fullmatch(targetTxHash) is not None)
        self.assertEqual(res["result"]["code"], 0)

        tx = await self.ain.getTransaction(targetTxHash)
        self.assertIsNone(tx)  # should be None

    @asyncTest
    async def test00SendSignedTransaction(self):
        tx = TransactionBody(
            nonce=-1,
            gas_price=500,
            timestamp=getTimestamp(),
            operation=SetOperation(
                type="SET_OWNER",
                ref=f"/apps/{APP_NAME}{PY_VERSION}",
                value={
                    ".owner": {
                        "owners": {
                            "*": {
                                "write_owner": True,
                                "write_rule": True,
                                "write_function": True,
                                "branch_owner": True
                            }
                        }
                    }
                }
            )
        )

        sig = self.ain.wallet.signTransaction(tx)
        res = await self.ain.sendSignedTransaction(sig, tx)
        targetTxHash = res["tx_hash"]
        self.assertFalse("code" in res)
        self.assertTrue(TX_PATTERN.fullmatch(targetTxHash) is not None)
        self.assertEqual(res["result"]["code"], 0)

    @asyncTest
    async def test00SendSignedTransactionInvalidSignature(self):
        tx = TransactionBody(
            nonce=-1,
            gas_price=500,
            timestamp=getTimestamp(),
            operation=SetOperation(
                type="SET_OWNER",
                ref=f"/apps/{APP_NAME}{PY_VERSION}",
                value={
                    ".owner": {
                        "owners": {
                            "*": {
                                "write_owner": True,
                                "write_rule": True,
                                "write_function": True,
                                "branch_owner": True
                            }
                        }
                    }
                }
            )
        )

        sig = ""
        raised = False
        try:
            await self.ain.sendSignedTransaction(sig, tx)
        except BlockchainError as e:
            self.assertEqual(e.code, 30302)
            self.assertEqual(e.message, "Missing properties.")
            raised = True
        self.assertTrue(raised)

    @asyncTest
    async def test01SendTransactionBatch(self):
        tx1 = TransactionInput(
            operation=SetOperation(
                type="SET_VALUE",
                ref=f"/apps/{APP_NAME}{PY_VERSION}/users",
                value={ self.defaultAddr: True }
            ),
            address=self.addr1
        )

        tx2 = TransactionInput(
            operation=SetMultiOperation(
                type="SET",
                op_list=[
                    SetOperation(
                        type="SET_OWNER",
                        ref=f"/apps/{APP_NAME}{PY_VERSION}/users",
                        value={
                            ".owner": {
                                "owners": {
                                    "*": {
                                        "write_owner": False,
                                        "write_rule": False,
                                        "write_function": True,
                                        "branch_owner": True
                                    }
                                }
                            }
                        }
                    ),
                    SetOperation(
                        type="SET_OWNER",
                        ref=f"/apps/{APP_NAME}{PY_VERSION}/users/{self.defaultAddr}",
                        value={
                            ".owner": {
                                "owners": {
                                    self.defaultAddr: {
                                        "write_owner": True,
                                        "write_rule": True,
                                        "write_function": True,
                                        "branch_owner": True
                                    }
                                }
                            }
                        }
                    )
                ]
            ),
            address=self.addr2
        )

        tx3 = TransactionInput(
            operation=SetOperation(
                type="SET_RULE",
                ref=f"/apps/{APP_NAME}{PY_VERSION}/users/{self.defaultAddr}",
                value={ ".rule": { "write": f'auth.addr === "{self.defaultAddr}"' } }
            )
        )

        tx4 = TransactionInput(
            operation=SetOperation(
                type="SET_VALUE",
                ref=f"/apps/{APP_NAME}{PY_VERSION}/users/{self.defaultAddr}",
                value=False
            )
        )

        tx5 = TransactionInput(
            operation=SetOperation(
                type="SET_VALUE",
                ref=f"/apps/{APP_NAME}{PY_VERSION}/users/{self.defaultAddr}",
                value=True
            ),
            address=self.addr1
        )

        tx6 = TransactionInput(
            operation=SetOperation(
                type="SET_RULE",
                ref=f"/apps/{APP_NAME}{PY_VERSION}/users/{self.defaultAddr}",
                value={ ".rule": { "write": "true" } }
            ),
            address=self.addr2
        )

        resps = await self.ain.sendTransactionBatch([tx1, tx2, tx3, tx4, tx5, tx6])
        for res in resps:
            self.assertTrue(TX_PATTERN.fullmatch(res["tx_hash"]) is not None)
        self.assertEqual(resps[0]["result"]["code"], 12103)
        self.assertEqual(resps[1]["result"]["result_list"]["0"]["code"], 0)
        self.assertEqual(resps[1]["result"]["result_list"]["1"]["code"], 0)
        self.assertEqual(resps[2]["result"]["code"], 0)
        self.assertEqual(resps[3]["result"]["code"], 0)
        self.assertEqual(resps[4]["result"]["code"], 12103)
        self.assertEqual(resps[5]["result"]["code"], 12302)

    @asyncTest
    async def test01SendTransactionBatchWithEmptyList(self):
        raised = False
        try:
            await self.ain.sendTransactionBatch([])
        except BlockchainError as e:
            self.assertEqual(e.code, 30401)
            self.assertEqual(e.message, "Invalid batch transaction format.")
            raised = True
        self.assertTrue(raised)

class TestDatabase(SnapshotTestCase):
    ain: Ain
    defaultAddr: str
    allowedPath: str

    @classmethod
    def setUpClass(cls):
        cls.ain = Ain(testNode, 0)
        cls.ain.wallet.addAndSetDefaultAccount(accountSk)
        cls.defaultAddr = cls.ain.wallet.defaultAccount.address
        cls.allowedPath = f"apps/{APP_NAME}{PY_VERSION}/users/{cls.defaultAddr}"
        super(TestDatabase, cls).setUpClass()

    def matchSnapshot(self, value):
        dump = json.dumps(
            value,
            default=lambda o: o.__dict__,
            separators=(",", ":"),
            sort_keys=True
        ).replace(f'{APP_NAME}{PY_VERSION}', APP_NAME)
        self.assertMatchSnapshot(json.loads(dump))

    def testRef(self):
        self.assertEqual(self.ain.db.ref().path, "/")
        self.assertEqual(self.ain.db.ref(self.allowedPath).path, "/" + self.allowedPath)
    
    @asyncTest
    async def test00SetOwnerIsDryrunTrue(self):
        res = await self.ain.db.ref(self.allowedPath).setOwner(ValueOnlyTransactionInput(
            value={
                ".owner": {
                    "owners": {
                        "*": {
                            "write_owner": True,
                            "write_rule": True,
                            "write_function": True,
                            "branch_owner": True
                        }
                    }
                }
            }
        ),
        True)
        self.assertEqual(res["result"]["code"], 0)
        self.assertEqual(res["result"]["is_dryrun"], True)

        fail = await self.ain.db.ref("/consensus").setOwner(ValueOnlyTransactionInput(
            value={
                ".owner": {
                    "owners": {
                        "*": {
                            "write_owner": True,
                            "write_rule": True,
                            "write_function": True,
                            "branch_owner": True
                        }
                    }
                }
            }
        ),
        True)
        self.assertEqual(fail["result"]["code"], 12501)
        self.assertEqual(fail["result"]["is_dryrun"], True)

    @asyncTest
    async def test00SetOwner(self):
        res = await self.ain.db.ref(self.allowedPath).setOwner(ValueOnlyTransactionInput(
            value={
                ".owner": {
                    "owners": {
                        "*": {
                            "write_owner": True,
                            "write_rule": True,
                            "write_function": True,
                            "branch_owner": True
                        }
                    }
                }
            }
        ))
        self.assertEqual(res["result"]["code"], 0)

        fail = await self.ain.db.ref("/consensus").setOwner(ValueOnlyTransactionInput(
            value={
                ".owner": {
                    "owners": {
                        "*": {
                            "write_owner": True,
                            "write_rule": True,
                            "write_function": True,
                            "branch_owner": True
                        }
                    }
                }
            }
        ))
        self.assertEqual(fail["result"]["code"], 12501)

    @asyncTest
    async def test00SetRuleIsDryrunTrue(self):
        res = await self.ain.db.ref(self.allowedPath).setRule(ValueOnlyTransactionInput(
            value={ ".rule": { "write": "true" } }
        ),
        True)
        self.assertEqual(res["result"]["code"], 0)
        self.assertEqual(res["result"]["is_dryrun"], True)

    @asyncTest
    async def test00SetRuleIsDryrunTrue(self):
        res = await self.ain.db.ref(self.allowedPath).setRule(ValueOnlyTransactionInput(
            value={ ".rule": { "write": "true" } }
        ),
        True)
        self.assertEqual(res["result"]["code"], 0)
        self.assertEqual(res["result"]["is_dryrun"], True)

    @asyncTest
    async def test00SetRule(self):
        res = await self.ain.db.ref(self.allowedPath).setRule(ValueOnlyTransactionInput(
            value={ ".rule": { "write": "true" } }
        ))
        self.assertEqual(res["result"]["code"], 0)

    @asyncTest
    async def test00SetValueIsDryrunTrue(self):
        res = await self.ain.db.ref(f"{self.allowedPath}/username").setValue(ValueOnlyTransactionInput(
            value="test_user"
        ),
        True)
        self.assertEqual(res["result"]["code"], 0)
        self.assertEqual(res["result"]["is_dryrun"], True)

    @asyncTest
    async def test00SetValue(self):
        res = await self.ain.db.ref(f"{self.allowedPath}/username").setValue(ValueOnlyTransactionInput(
            value="test_user"
        ))
        self.assertEqual(res["result"]["code"], 0)

    @asyncTest
    async def test00SetFunctionIsDryrunTrue(self):
        res = await self.ain.db.ref(self.allowedPath).setFunction(ValueOnlyTransactionInput(
            value={
                ".function": {
                    "0xFUNCTION_HASH": {
                        "function_url": "https://events.ainetwork.ai/trigger",
                        "function_id": "0xFUNCTION_HASH",
                        "function_type": "REST"
                    }
                }
            }
        ),
        True)
        self.assertEqual(res["result"]["code"], 0)
        self.assertEqual(res["result"]["is_dryrun"], True)

    @asyncTest
    async def test00SetFunction(self):
        res = await self.ain.db.ref(self.allowedPath).setFunction(ValueOnlyTransactionInput(
            value={
                ".function": {
                    "0xFUNCTION_HASH": {
                        "function_url": "https://events.ainetwork.ai/trigger",
                        "function_id": "0xFUNCTION_HASH",
                        "function_type": "REST"
                    }
                }
            }
        ))
        self.assertEqual(res["result"]["code"], 0)

    @asyncTest
    async def test00SetIsDryrunTrue(self):
        res = await self.ain.db.ref(self.allowedPath).set(SetMultiTransactionInput(
            op_list=[
                SetOperation(
                    type="SET_RULE",
                    ref="can/write/",
                    value={ ".rule": { "write": "true" } }
                ),
                SetOperation(
                    type="SET_RULE",
                    ref="cannot/write",
                    value={ ".rule": { "write": "false" } }
                ),
                SetOperation(
                    type="INC_VALUE",
                    ref="can/write/",
                    value=5
                ),
                SetOperation(
                    type="DEC_VALUE",
                    ref="can/write",
                    value=10
                )
            ],
            nonce=-1
        ),
        True)
        self.assertEqual(len(res["result"]["result_list"].keys()), 4)
        self.assertEqual(res["result"]["is_dryrun"], True)

    @asyncTest
    async def test00Set(self):
        res = await self.ain.db.ref(self.allowedPath).set(SetMultiTransactionInput(
            op_list=[
                SetOperation(
                    type="SET_RULE",
                    ref="can/write/",
                    value={ ".rule": { "write": "true" } }
                ),
                SetOperation(
                    type="SET_RULE",
                    ref="cannot/write",
                    value={ ".rule": { "write": "false" } }
                ),
                SetOperation(
                    type="INC_VALUE",
                    ref="can/write/",
                    value=5
                ),
                SetOperation(
                    type="DEC_VALUE",
                    ref="can/write",
                    value=10
                )
            ],
            nonce=-1
        ))
        self.assertEqual(len(res["result"]["result_list"].keys()), 4)

    @asyncTest
    async def test01GetValue(self):
        self.matchSnapshot(await self.ain.db.ref(self.allowedPath).getValue())
    
    @asyncTest
    async def test01GetRule(self):
        self.matchSnapshot(await self.ain.db.ref(self.allowedPath).getRule())
    
    @asyncTest
    async def test01GetOwner(self):
        self.matchSnapshot(await self.ain.db.ref(self.allowedPath).getOwner())
    
    @asyncTest
    async def test01GetFunction(self):
        self.matchSnapshot(await self.ain.db.ref(self.allowedPath).getFunction())

    @asyncTest
    async def test01Get(self):
        self.matchSnapshot(await self.ain.db.ref(self.allowedPath).get([
            GetOperation(
                type="GET_RULE",
                ref=""
            ),
            GetOperation(
                type="GET_VALUE"
            ),
            GetOperation(
                type="GET_VALUE",
                ref="deeper/path"
            )
        ]))

    @asyncTest
    async def test01GetWithEmptyList(self):
        raised = False
        try:
            await self.ain.db.ref(self.allowedPath).get([])
        except BlockchainError as e:
            self.assertEqual(e.code, 30006)
            self.assertEqual(e.message, "Invalid op_list given")
            raised = True
        self.assertTrue(raised)

    @asyncTest
    async def test01GetWithOptions(self):
        self.matchSnapshot(await self.ain.db.ref().getValue(self.allowedPath, GetOptions(is_final=True)))
        self.matchSnapshot(await self.ain.db.ref().getValue(self.allowedPath, GetOptions(is_global=True)))
        self.matchSnapshot(await self.ain.db.ref().getValue(self.allowedPath, GetOptions(is_shallow=True)))
        self.matchSnapshot(await self.ain.db.ref().getValue(self.allowedPath, GetOptions(include_proof=True)))
        self.matchSnapshot(await self.ain.db.ref().getValue(self.allowedPath, GetOptions(include_tree_info=True)))
        getWithVersion = await self.ain.db.ref().getValue(self.allowedPath, GetOptions(include_version=True))
        self.assertTrue("#version" in getWithVersion)

    @asyncTest
    async def test02DeleteValueIsDryrunTrue(self):
        res = await self.ain.db.ref(f"{self.allowedPath}/can/write").deleteValue(isDryrun=True)
        self.assertEqual(res["result"]["code"], 0)
        self.assertEqual(res["result"]["is_dryrun"], True)

    @asyncTest
    async def test02DeleteValue(self):
        res = await self.ain.db.ref(f"{self.allowedPath}/can/write").deleteValue()
        self.assertEqual(res["result"]["code"], 0)

    @asyncTest
    async def test03EvalRuleTrue(self):
        res = await self.ain.db.ref(self.allowedPath).evalRule(EvalRuleInput(
            ref="/can/write",
            value="hi"
        ))
        self.assertEqual(res["code"], 0)
        self.matchSnapshot(res)

    @asyncTest
    async def test03EvalRuleFalse(self):
        res = await self.ain.db.ref(self.allowedPath).evalRule(EvalRuleInput(
            ref="/cannot/write",
            value="hi"
        ))
        self.assertEqual(res["code"], 12103)

    @asyncTest
    async def test03EvalOwner(self):
        self.matchSnapshot(await self.ain.db.ref(self.allowedPath).evalOwner(EvalOwnerInput(
            permission="branch_owner"
        )))

    @asyncTest
    async def test03MatchFunction(self):
        self.matchSnapshot(await self.ain.db.ref(self.allowedPath).matchFunction())
    
    @asyncTest
    async def test03MatchRule(self):
        self.matchSnapshot(await self.ain.db.ref(self.allowedPath).matchRule())
    
    @asyncTest
    async def test03MatchOwner(self):
        self.matchSnapshot(await self.ain.db.ref(self.allowedPath).matchOwner())
    



from typing import TYPE_CHECKING, Any, List, Optional
from ain.types import (
    SetOperation,
    GetOperation,
    SetOperationType,
    GetOperationType,
    SetMultiOperationType,
    SetMultiOperation,
    TransactionInput,
    ValueOnlyTransactionInput,
    SetMultiTransactionInput,
    EvalRuleInput,
    EvalOwnerInput,
    MatchInput,
    GetOptions,
)
from ain.db.push_id import PushId

if TYPE_CHECKING:
    from ain.ain import Ain

class Reference:
    """Class that creates a reference to a path in the AIN Database."""

    _path: str
    _key: Optional[str]
    _ain: "Ain"
    _pushId: PushId
    _isGlobal: bool

    def __init__(self, ain: "Ain", path: str = None):
        self._path = Reference.sanitizeRef(path)
        pathArr = self._path.split("/")
        if len(pathArr) > 0:
            self._key = pathArr[-1]
        else:
            self._key = None
        self._ain = ain
        self._pushId = PushId()
        self._isGlobal = False

    @property
    def path(self) -> str:
        """The base path of this reference."""
        return self._path

    @property
    def key(self) -> Optional[str]:
        """The key of this reference."""
        return self._key

    def setIsGlobal(self, isGlobal: bool):
        """Sets a global reference using flag.

        Args:
            isGlobal (bool): The global reference using flag.
        """
        self._isGlobal = isGlobal

    def push(self) -> "Reference":
        """Creates a key for a new child of base path virtually.
        So doesn't set any values on the AIN Blockchain.

        Returns:
            Reference: The instance of `Reference` of a new child.
        """
        newKey = "/" + self._pushId.generate()
        return Reference(self._ain, Reference.extendPath(self._path, newKey))

    async def pushValue(self, value: Any = None) -> Any:
        """Sets the value at a new child of base path.

        Args:
            value (Any, Optional): The value that you want to set. Defaults to `None`.
        
        Returns:
            The result of the transaction.
        """
        return await self.push().setValue(ValueOnlyTransactionInput(value=value))

    async def getValue(self, path: str = None, options: GetOptions = None) -> Any:
        """Gets a value at the `path`.

        Args:
            path (str, Optional): The path that you want to extend under the base path.
                If `path` is `None`, no extending. Defaults to `None`.
            options (GetOptions, Optional): The options for this transaction.
                Defaults to no options.

        Returns:
            The value at the `path`.
        """
        req = self.buildGetRequest("GET_VALUE", Reference.extendPath(self.path, path), options)
        return await self._ain.provider.send("ain_get", req)

    async def getRule(self, path: str = None, options: GetOptions = None) -> Any:
        """Gets a rule at the `path`.

        Args:
            path (str, Optional): The path that you want to extend under the base path.
                If `path` is `None`, no extending. Defaults to `None`.
            options (GetOptions, Optional): The options for this transaction.
                Defaults to no options.

        Returns:
            The rule at the `path`.
        """
        req = self.buildGetRequest("GET_RULE", Reference.extendPath(self.path, path), options)
        return await self._ain.provider.send("ain_get", req)

    async def getOwner(self, path: str = None, options: GetOptions = None) -> Any:
        """Gets an owner config at the `path`.

        Args:
            path (str, Optional): The path that you want to extend under the base path.
                If `path` is `None`, no extending. Defaults to `None`.
            options (GetOptions, Optional): The options for this transaction.
                Defaults to no options.
        
        Returns:
            The owner config at the `path`.
        """
        req = self.buildGetRequest("GET_OWNER", Reference.extendPath(self.path, path), options)
        return await self._ain.provider.send("ain_get", req)

    async def getFunction(self, path: str = None, options: GetOptions = None) -> Any:
        """Gets a function config at the `path`.

        Args:
            path (str, Optional): The path that you want to extend under the base path.
                If `path` is `None`, no extending. Defaults to `None`.
            options (GetOptions, Optional): The options for this transaction.
                Defaults to no options.
        
        Returns:
            The function config at the `path`.
        """
        req = self.buildGetRequest("GET_FUNCTION", Reference.extendPath(self.path, path), options)
        return await self._ain.provider.send("ain_get", req)

    async def get(self, gets: List[GetOperation]) -> Any:
        """Gets a value, write rule, owner rule, or function hash at multiple paths.

        Args:
            gets (List[GetOperation]): The array of the get requests.
                Could be any one from "VALUE", "RULE", "OWNER", "FUNC" or a combination of them as an array.

        Returns:
            The value, write rule, owner rule, or function hash at multiple paths.
        """
        extendedGets: List[GetOperation] = []
        for get in gets:
            extendedGets.append(
                GetOperation(
                    type=get.type,
                    ref=Reference.extendPath(self._path, getattr(get, "ref", None)),
                    is_global=getattr(get, "is_global", None),
                )
            )
        req = {"type": "GET", "op_list": extendedGets}
        return await self._ain.provider.send("ain_get", req)

    async def deleteValue(self, transactionInput: ValueOnlyTransactionInput = None, isDryrun = False) -> Any:
        """Deletes the value.

        Args:
            transactionInput (ValueOnlyTransactionInput): The transaction input object.
                Any value given will be overwritten with null.
            isDryrun (bool): Dryrun option.

        Returns:
            The result of the transaction.
        """
        txInput = ValueOnlyTransactionInput()
        if transactionInput is not None:
            txInput = transactionInput
        ref = Reference.extendPath(self._path, getattr(txInput, "ref", None))
        operation = SetOperation(
            type="SET_VALUE",
            ref=ref,
            is_global=getattr(txInput, "is_global", self._isGlobal),
        )
        operation.value = None
        return await self._ain.sendTransaction(
            TransactionInput(
                operation=operation,
                parent_tx_hash=getattr(txInput, "parent_tx_hash", None),
            ),
            isDryrun
        )

    async def setFunction(self, transactionInput: ValueOnlyTransactionInput, isDryrun = False) -> Any:
        """Sets the function config.

        Args:
            transactionInput (ValueOnlyTransactionInput): The transaction input object.
            isDryrun (bool): Dryrun option.

        Returns:
            The result of the transaction.
        """
        ref = Reference.extendPath(self._path, getattr(transactionInput, "ref", None))
        return await self._ain.sendTransaction(
            Reference.extendSetTransactionInput(
                transactionInput,
                ref,
                "SET_FUNCTION",
                self._isGlobal
            ),
            isDryrun
        )

    async def setOwner(self, transactionInput: ValueOnlyTransactionInput, isDryrun = False) -> Any:
        """Sets the owner rule.

        Args:
            transactionInput (ValueOnlyTransactionInput): The transaction input object.
            isDryrun (bool): Dryrun option.
            
        Returns:
            The result of the transaction.
        """
        ref = Reference.extendPath(self._path, getattr(transactionInput, "ref", None))
        return await self._ain.sendTransaction(
            Reference.extendSetTransactionInput(
                transactionInput,
                ref,
                "SET_OWNER",
                self._isGlobal
            ),
            isDryrun
        )

    async def setRule(self, transactionInput: ValueOnlyTransactionInput, isDryrun = False) -> Any:
        """Sets the write rule.

        Args:
            transactionInput (ValueOnlyTransactionInput): The transaction input object.
            isDryrun (bool): Dryrun option.
            
        Returns:
            The result of the transaction.
        """
        ref = Reference.extendPath(self._path, getattr(transactionInput, "ref", None))
        return await self._ain.sendTransaction(
            Reference.extendSetTransactionInput(
                transactionInput,
                ref,
                "SET_RULE",
                self._isGlobal
            ),
            isDryrun
        )

    async def setValue(self, transactionInput: ValueOnlyTransactionInput, isDryrun = False) -> Any:
        """Sets the value.

        Args:
            transactionInput (ValueOnlyTransactionInput): The transaction input object.
            isDryrun (bool): Dryrun option.
            
        Returns:
            The result of the transaction.
        """
        ref = Reference.extendPath(self._path, getattr(transactionInput, "ref", None))
        return await self._ain.sendTransaction(
            Reference.extendSetTransactionInput(
                transactionInput,
                ref,
                "SET_VALUE",
                self._isGlobal
            ),
            isDryrun
        )

    async def incrementValue(self, transactionInput: ValueOnlyTransactionInput, isDryrun = False) -> Any:
        """Increments the value.

        Args:
            transactionInput (ValueOnlyTransactionInput): The transaction input object.
            isDryrun (bool): Dryrun option.
            
        Returns:
            The result of the transaction.
        """
        ref = Reference.extendPath(self._path, getattr(transactionInput, "ref", None))
        return await self._ain.sendTransaction(
            Reference.extendSetTransactionInput(
                transactionInput,
                ref,
                "INC_VALUE",
                self._isGlobal
            ),
            isDryrun
        )

    async def decrementValue(self, transactionInput: ValueOnlyTransactionInput, isDryrun = False) -> Any:
        """Decrements the value.
        
        Args:
            transactionInput (ValueOnlyTransactionInput): The transaction input object.
            isDryrun (bool): Dryrun option.
            
        Returns:
            The result of the transaction.
        """
        ref = Reference.extendPath(self._path, getattr(transactionInput, "ref", None))
        return await self._ain.sendTransaction(
            Reference.extendSetTransactionInput(
                transactionInput,
                ref,
                "DEC_VALUE",
                self._isGlobal
            ),
            isDryrun
        )

    async def set(self, transactionInput: SetMultiTransactionInput, isDryrun = False) -> Any:
        """Processes the multiple set operations.
        
        Args:
            transactionInput (SetMultiTransactionInput): The transaction input object.
            isDryrun (bool): Dryrun option.
            
        Returns:
            The result of the transaction.
        """
        return await self._ain.sendTransaction(
            Reference.extendSetMultiTransactionInput(
                transactionInput,
                self._path,
            ),
            isDryrun
        )

    async def evalRule(self, params: EvalRuleInput) -> Any:
        """Evals the rule evaluation result.

        Args:
            params (EvalRuleInput): The eval rule input object.
            
        Returns:
            The result of the evaluation.
            `True`, if the params satisfy the write rule,
            `False,` if not.
        """
        address = self._ain.signer.getAddress(getattr(params, "address", None))
        req = {
            "address": address,
            "ref": Reference.extendPath(self._path, getattr(params, "ref", None)),
            "value": params.value,
        }
        if hasattr(params, "timestamp"):
            req["timestamp"] = params.timestamp
        return await self._ain.provider.send("ain_evalRule", req)

    async def evalOwner(self, params: EvalOwnerInput) -> Any:
        """Evals the owner evaluation result.

        Args:
            params (EvalOwnerInput): The eval owner input object.
            
        Returns:
            The owner evaluation result.
        """
        address = self._ain.signer.getAddress(getattr(params, "address", None))
        req = {
            "address": address,
            "ref": Reference.extendPath(self._path, getattr(params, "ref", None)),
            "permission": params.permission,
        }
        return await self._ain.provider.send("ain_evalOwner", req)

    async def matchFunction(self, params: MatchInput = None) -> Any:
        """Matches the function configs that are related to the input ref.

        Args:
            params (MatchInput): The match input object.
            
        Returns:
            The function configs that are related to the input ref.
        """
        ref = self._path
        if params is not None and hasattr(params, "ref"):
            ref = Reference.extendPath(ref, getattr(params, "ref", None))
        return await self._ain.provider.send("ain_matchFunction", {"ref": ref})

    async def matchRule(self, params: MatchInput = None) -> Any:
        """Matches the rule configs that are related to the input ref.

        Args:
            params (MatchInput): The match input object.
            
        Returns:
            The rule configs that are related to the input ref.
        """
        ref = self._path
        if params is not None and hasattr(params, "ref"):
            ref = Reference.extendPath(ref, getattr(params, "ref", None))
        return await self._ain.provider.send("ain_matchRule", {"ref": ref})

    async def matchOwner(self, params: MatchInput = None) -> Any:
        """Matches the owner configs that are related to the input ref.

        Args:
            params (MatchInput): The match input object.
            
        Returns:
            The owner configs that are related to the input ref.
        """
        ref = self._path
        if params is not None and hasattr(params, "ref"):
            ref = Reference.extendPath(ref, getattr(params, "ref", None))
        return await self._ain.provider.send("ain_matchOwner", {"ref": ref})

    @staticmethod
    def buildGetRequest(type: GetOperationType, ref: str, options: GetOptions = None) -> dict:
        """Builds a get request.

        Args:
            type (GetOperationType): The type of get operation.
            ref (str): The path that you want to make transaction.
            options (GetOptions, Optional): The options for this transaction.
                Defaults to no options.
        
        Returns:
            dict: The builded get request.
        """
        request = {"type": type, "ref": Reference.sanitizeRef(ref)}
        if options is not None:
            request.update(options.__dict__)
        return request

    @staticmethod
    def extendPath(basePath: str = None, extension: str = None) -> str:
        """Extends the path.

        Args:
            basePath (str, Optional): The base path that you want.
            extension (str, Optional): The extension that you want to extend under the base path.
        
        Returns:
            str: The extended `basePath`, by the `extension`.
        """
        sanitizedBase = Reference.sanitizeRef(basePath)
        sanitizedExt = Reference.sanitizeRef(extension)
        if sanitizedBase == "/":
            return sanitizedExt
        if sanitizedExt == "/":
            return sanitizedBase
        return sanitizedBase + sanitizedExt

    @staticmethod
    def extendSetTransactionInput(
        input: ValueOnlyTransactionInput,
        ref: str,
        type: SetOperationType, 
        isGlobal: bool
    ) -> TransactionInput:
        """Extends the transaction input with an appropriate type, ref and value.

        Args:
            input (ValueOnlyTransactionInput): The transaction input object.
            ref (str): The path that you want to make transaction.
            type (SetOperationType): The type of set operation.
        
        Returns:
            TransactionInput: The decorated transaction input.
        """
        operation = SetOperation(
            type=type,
            ref=ref,
            value=getattr(input, "value", None),
            is_global=getattr(input, "is_global", isGlobal),
        )
        return TransactionInput(
            operation=operation,
            parent_tx_hash=getattr(input, "parent_tx_hash", None),
            nonce=getattr(input, "nonce", None),
            address=getattr(input, "address", None),
            timestamp=getattr(input, "timestamp", None),
            gas_price=getattr(input, "gas_price", None),
        )
        
    @staticmethod
    def extendSetMultiTransactionInput(
        input: SetMultiTransactionInput,
        ref: str,
        type: SetMultiOperationType = "SET",
    ) -> TransactionInput:
        """Extends the transaction input with an appropriate type and op_list.

        Args:
            input (SetMultiTransactionInput): The transaction input object.
            ref (str): The path that you want to make transaction.
            type (SetMultiOperationType): The type of set operations.
        
        Returns:
            TransactionInput: The decorated transaction input.
        """
        op_list: List[SetOperation] = []
        for op in input.op_list:
            op_list.append(
                SetOperation(
                    type=op.type,
                    ref=Reference.extendPath(ref, op.ref),
                    value=getattr(op, "value", None),
                    is_global=getattr(op, "is_global", None)
                )
            )
            
        operation = SetMultiOperation(type, op_list)
        return TransactionInput(
            operation=operation,
            parent_tx_hash=getattr(input, "parent_tx_hash", None),
            nonce=getattr(input, "nonce", None),
            address=getattr(input, "address", None),
            timestamp=getattr(input, "timestamp", None),
            gas_price=getattr(input, "gas_price", None),
        )

    @staticmethod
    def sanitizeRef(ref: str = None) -> str:
        """Sanitizes the path.

        Args:
            ref (str): A path that you want to sanitize.
        
        Returns:
            str: A sanitized ref. It should have a slash at the beginning and no slash at the end.
        """
        if ref is None:
            return "/"
        return "/" + "/".join(filter(lambda s: s != "", ref.split("/")))



from unittest import TestCase
from ain.account import Account
from ain.utils import *
from .data import (
    mnemonic,
    mnemonicPrivateKey,
    mnemonicPublicKey,
    mnemonicAddress,
)

class TestAccount(TestCase):
    def testAccountFromMnemonic(self):
        account = Account.fromMnemonic(mnemonic)
        self.assertEqual(account.private_key, mnemonicPrivateKey.hex())
        self.assertEqual(account.public_key, mnemonicPublicKey.hex())
        self.assertEqual(account.address, mnemonicAddress)
    
    def testAccountCreateFromEntropyNone(self):
        account = Account.create()
        publicKey = privateToPublic(bytes.fromhex(account.private_key)).hex()
        self.assertEqual(account.public_key, publicKey)

    def testAccountCreateFromEntropy(self):
        account = Account.create("Test")
        publicKey = privateToPublic(bytes.fromhex(account.private_key)).hex()
        self.assertEqual(account.public_key, publicKey)


from abc import *
import asyncio
from typing import List, Any
from ain.wallet import Wallet
from ain.provider import Provider
from ain.signer import Signer
from ain.types import TransactionInput, TransactionBody
from ain.utils import getTimestamp, toChecksumAddress

class DefaultSigner(Signer):
    """The default concrete class of Signer abstract class implemented using Wallet class.
       When Ain class is initialized, DefaultSigner is set as its signer.

    Args:
        wallet (Wallet): The wallet to initialize with.
        provider (Provider): The network provider to initialize with.
    """
    
    wallet: Wallet
    """The `Wallet` instance."""
    provider: Provider
    """The `Provider` instance."""

    def __init__(self, wallet: Wallet, provider: Provider):
        self.wallet = wallet
        self.provider = provider

    def getAddress(self, address: str = None) -> str:
        """Gets an account's checksum address.
           If the address is not given, the default account of the signer is used.

        Args:
            address (str, Optional): The address of the account.

        Returns:
            str: The checksum address.
        """
        return self.wallet.getImpliedAddress(address)

    async def signMessage(self, message: Any, address: str = None) -> str:
        """Signs a message using an account.
           If an address is not given, the default account of the signer is used.

        Args:
            message (Any): The message to sign.
            address (str, Optional): The address of the account.

        Returns:
            str: The signature.
        """
        return self.wallet.sign(message, address)

    async def sendTransaction(self, transactionObject: TransactionInput, isDryrun = False) -> Any:
        """Signs and sends a transaction to the network.

        Args:
            transactionObject (TransactionInput): The transaction input object.
            isDryrun (bool): The dryrun option.

        Returns:
            The return value of the blockchain API.
        """
        txBody = await self.buildTransactionBody(transactionObject)
        signature = self.wallet.signTransaction(txBody, getattr(transactionObject, "address", None))
        return await self.sendSignedTransaction(signature, txBody, isDryrun)

    async def sendTransactionBatch(self, transactionObjects: List[TransactionInput]) -> List[Any]:
        """Signs and sends multiple transactions in a batch to the network.

        Args:
            transactionObjects (List[TransactionInput]): The list of the transaction input objects.
        
        Returns:
            The return value of the blockchain API.
        """
        txListCoroutines = []
        for txInput in transactionObjects:
            txListCoroutines.append(self.buildTransactionBodyAndSignature(txInput))
        
        txList = await asyncio.gather(*txListCoroutines)
        return await self.provider.send("ain_sendSignedTransactionBatch", {"tx_list": txList})

    async def sendSignedTransaction(self, signature: str, txBody: TransactionBody, isDryrun = False) -> Any:
        """Sends a signed transaction to the network.

        Args:
            signature (str): The signature.
            txBody (TransactionBody): The transaction body.
            isDryrun (bool): The dryrun option.

        Returns:
            The return value of the blockchain API.
        """
        method = "ain_sendSignedTransactionDryrun" if isDryrun == True else "ain_sendSignedTransaction"
        return await self.provider.send(method, {"signature": signature, "tx_body": txBody})

    async def buildTransactionBody(self, transactionInput: TransactionInput) -> TransactionBody:
        """Builds a transaction body object from a transaction input object.

        Args:
            transactionInput (TransactionInput): The transaction input object.
        
        Returns:
            The TransactionBody object.
        """
        address = self.getAddress(
            getattr(transactionInput, "address", None)
        )
        operation = transactionInput.operation
        parent_tx_hash = getattr(transactionInput, "parent_tx_hash", None)
        nonce = getattr(transactionInput, "nonce", None)
        if nonce is None:
            nonce = await self.getNonce({"address": address, "from": "pending"})
        timestamp = getattr(transactionInput, "timestamp", getTimestamp())
        gas_price = getattr(transactionInput, "gas_price", 0)
        billing = getattr(transactionInput, "billing", None)

        return TransactionBody(
            operation=operation,
            parent_tx_hash=parent_tx_hash,
            nonce=nonce,
            timestamp=timestamp,
            gas_price=gas_price,
            billing=billing,
        )

    async def buildTransactionBodyAndSignature(self, transactionObject: TransactionInput) -> dict:
        """Builds a transaction body object and a signature from a transaction input object.

        Args:
            transactionInput (TransactionInput): The transaction input object.
        
        Returns:
            The TransactionBody object and the signature.
        """
        txBody = await self.buildTransactionBody(transactionObject)
        if not hasattr(transactionObject, "nonce"):
            # Batch transactions' nonces should be specified.
            # If they're not, they default to un-nonced (nonce = -1).
            txBody.nonce = -1
            
        signature = self.wallet.signTransaction(txBody, getattr(transactionObject, "address", None))
        return {"signature": signature, "tx_body": txBody}

    async def getNonce(self, args: dict) -> Any:
        """Fetches an account's nonce value, which is the current transaction count of the account.

        Args:
            args (dict): The ferch options.
                It may contain a string 'address' value and a string 'from' value.
                The 'address' is the address of the account to get the nonce of,
                and the 'from' is the source of the data.
                It could be either the pending transaction pool ("pending") or
                the committed blocks ("committed"). The default value is "committed".
        
        Returns:
            The nonce value.
        """
        params = dict(args)
        if "address" in args:
            params["address"] = toChecksumAddress(args["address"])
        else:
            params["address"] = self.getAddress()

        if "from" in args:
            if args["from"] != "pending" and args["from"] != "committed":
                raise ValueError("'from' should be either 'pending' or 'committed'")

        return await self.provider.send("ain_getNonce", params)
    


from typing import TYPE_CHECKING
from ain.db.ref import Reference

if TYPE_CHECKING:
    from ain.ain import Ain
    from ain.provider import Provider

class Database:
    """Class for the AIN Database instance."""

    _ain: "Ain"
    _provider: "Provider"
    
    def __init__(self, ain: "Ain", provider: "Provider"):
        self._ain = ain
        self._provider = provider

    def ref(self, path: str = None) -> Reference:
        """Creates a `Reference` to the given path at the AIN Blockchain Database(global state tree).

        Args:
            path (str, Optional): The path that you want to create a `Reference`.
                If `path` is `None`, references the root of the tree. Defaults to `None`.

        Returns:
            Reference: The `Reference` at the path.
        """
        return Reference(self._ain, path)




from unittest import TestCase
from ain.utils import *
from ain.utils.v3keystore import *
from ain.types import SetOperation, TransactionBody
from .data import (
    address,
    pk,
    sk,
    mnemonic,
    mnemonicPrivateKey,
    checksumAddresses,
    message,
    correctSignature,
    tx,
    txDifferent,
    v3KeystoreKDFList,
    v3KeystoreCipherList,
    v3KeystorePassword,
    v3KeystoreJSONList,
    v3KeystoreFixedUUID,
    v3KeystoreFixedIV,
    v3KeystoreFixedSalt,
    v3KeystoreFixedJSON,
    encrypted,
)

class TestKeccak(TestCase):
    def testKeccak(self):
        msg = "0x3c9229289a6125f7fdf1885a77bb12c37a8d3b4962d936f7e3084dece32a3ca1"
        r = "82ff40c0a986c6a5cfad4ddf4c3aa6996f1a7837f9c398e17e5de5cbd5a12b28"
        hash = keccak(msg)
        self.assertEqual(hash.hex(), r)

    def testKeccakWithoutHexprefix(self):
        msg = "3c9229289a6125f7fdf1885a77bb12c37a8d3b4962d936f7e3084dece32a3ca1"
        r = "22ae1937ff93ec72c4d46ff3e854661e3363440acd6f6e4adf8f1a8978382251"
        hash = keccak(msg)
        self.assertEqual(hash.hex(), r)

class TestCountDecimals(TestCase):
    def testCountDecimals(self):
        self.assertEqual(countDecimals(0), 0)  # '0'
        self.assertEqual(countDecimals(1), 0)  # '1'
        self.assertEqual(countDecimals(10), 0)  # '10'
        self.assertEqual(countDecimals(100), 0)  # '100'
        self.assertEqual(countDecimals(1000), 0)  # '1000'
        self.assertEqual(countDecimals(10000), 0)  # '10000'
        self.assertEqual(countDecimals(100000), 0)  # '100000'
        self.assertEqual(countDecimals(1000000), 0)  # '1000000'
        self.assertEqual(countDecimals(10000000), 0)  # '10000000'
        self.assertEqual(countDecimals(100000000), 0)  # '100000000'
        self.assertEqual(countDecimals(1000000000), 0)  # '1000000000'
        self.assertEqual(countDecimals(1234567890), 0)  # '1234567890'
        self.assertEqual(countDecimals(-1), 0)  # '-1'
        self.assertEqual(countDecimals(-1000000000), 0)  # '-1000000000'
        self.assertEqual(countDecimals(11), 0)  # '11'
        self.assertEqual(countDecimals(101), 0)  # '101'
        self.assertEqual(countDecimals(1001), 0)  # '1001'
        self.assertEqual(countDecimals(10001), 0)  # '10001'
        self.assertEqual(countDecimals(100001), 0)  # '100001'
        self.assertEqual(countDecimals(1000001), 0)  # '1000001'
        self.assertEqual(countDecimals(10000001), 0)  # '10000001'
        self.assertEqual(countDecimals(100000001), 0)  # '100000001'
        self.assertEqual(countDecimals(1000000001), 0)  # '1000000001'
        self.assertEqual(countDecimals(-11), 0)  # '-11'
        self.assertEqual(countDecimals(-1000000001), 0)  # '-1000000001'
        self.assertEqual(countDecimals(0.1), 1)  # '0.1'
        self.assertEqual(countDecimals(0.01), 2)  # '0.01'
        self.assertEqual(countDecimals(0.001), 3)  # '0.001'
        self.assertEqual(countDecimals(0.0001), 4)  # '0.0001'
        self.assertEqual(countDecimals(0.00001), 5)  # '1e-05'
        self.assertEqual(countDecimals(0.000001), 6)  # '1e-06'
        self.assertEqual(countDecimals(0.0000001), 7)  # '1e-07'
        self.assertEqual(countDecimals(0.00000001), 8)  # '1e-08'
        self.assertEqual(countDecimals(0.000000001), 9)  # '1e-09'
        self.assertEqual(countDecimals(0.0000000001), 10)  # '1e-10'
        self.assertEqual(countDecimals(-0.1), 1)  # '-0.1'
        self.assertEqual(countDecimals(-0.0000000001), 10)  # '-1e-10'
        self.assertEqual(countDecimals(1.2), 1)  # '1.2'
        self.assertEqual(countDecimals(0.12), 2)  # '0.12'
        self.assertEqual(countDecimals(0.012), 3)  # '0.012'
        self.assertEqual(countDecimals(0.0012), 4)  # '0.0012'
        self.assertEqual(countDecimals(0.00012), 5)  # '0.00012'
        self.assertEqual(countDecimals(0.000012), 6)  # '1.2e-05'
        self.assertEqual(countDecimals(0.0000012), 7)  # '1.2e-06'
        self.assertEqual(countDecimals(0.00000012), 8)  # '1.2e-07'
        self.assertEqual(countDecimals(0.000000012), 9)  # '1.2e-08'
        self.assertEqual(countDecimals(0.0000000012), 10)  # '1.2e-09'
        self.assertEqual(countDecimals(0.00000000012), 11)  # '1.2e-10'
        self.assertEqual(countDecimals(-1.2), 1)  # '-1.2'
        self.assertEqual(countDecimals(-0.00000000012), 11)  # '-1.2e-10'
        self.assertEqual(countDecimals(1.03), 2)  # '1.03'
        self.assertEqual(countDecimals(1.003), 3)  # '1.003'
        self.assertEqual(countDecimals(1.0003), 4)  # '1.0003'
        self.assertEqual(countDecimals(1.00003), 5)  # '1.00003'
        self.assertEqual(countDecimals(1.000003), 6)  # '1.000003'
        self.assertEqual(countDecimals(1.0000003), 7)  # '1.0000003'
        self.assertEqual(countDecimals(1.00000003), 8)  # '1.00000003'
        self.assertEqual(countDecimals(1.000000003), 9)  # '1.000000003'
        self.assertEqual(countDecimals(1.0000000003), 10)  # '1.0000000003'
        self.assertEqual(countDecimals(-1.03), 2)  # '-1.03'
        self.assertEqual(countDecimals(-1.0000000003), 10)  # '-1.0000000003'

class TestToJsLikeFloats(TestCase):
    def testToJsLikeFloats(self):
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":0.0001},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":0.0001},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":1e-05},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":0.00001},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":1e-06},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":0.000001},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":1e-07},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":1e-07},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":0.00012},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":0.00012},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":1.2e-05},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":0.000012},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":1.2e-06},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":0.0000012},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":1.2e-07},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":1.2e-07},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-0.0001},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-0.0001},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-1e-05},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-0.00001},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-1e-06},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-0.000001},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-1e-07},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-1e-07},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-0.00012},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-0.00012},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-1.2e-05},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-0.000012},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-1.2e-06},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-0.0000012},"timestamp":123}')
        self.assertEqual(
            toJsLikeFloats('{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-1.2e-07},"timestamp":123}'),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-1.2e-07},"timestamp":123}')

class TestToJsonString(TestCase):
    def testToJsonString(self):
        txBody = TransactionBody(
            operation=SetOperation(
                ref="/afan",
                value=100,
                type="SET_VALUE",
            ),
            nonce=10,
            timestamp=123,
        )
        self.assertEqual(
            toJsonString(txBody),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":100},"timestamp":123}')

    def testToJsonStringWithFloatValueWithoutExponent(self):
        txBody = TransactionBody(
            operation=SetOperation(
                ref="/afan",
                value=1.000003,  # '1.000003'
                type="SET_VALUE",
            ),
            nonce=10,
            timestamp=123,
        )
        self.assertEqual(
            toJsonString(txBody),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":1.000003},"timestamp":123}')

    def testToJsonStringWithFloatValueWithExponent4(self):
        txBody = TransactionBody(
            operation=SetOperation(
                ref="/afan",
                value=0.00012,  # '0.00012'
                type="SET_VALUE",
            ),
            nonce=10,
            timestamp=123,
        )
        self.assertEqual(
            toJsonString(txBody),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":0.00012},"timestamp":123}')  # 0.00012

    def testToJsonStringWithFloatValueWithExponent5(self):
        txBody = TransactionBody(
            operation=SetOperation(
                ref="/afan",
                value=0.000012,  # '1.2e-05'
                type="SET_VALUE",
            ),
            nonce=10,
            timestamp=123,
        )
        self.assertEqual(
            toJsonString(txBody),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":0.000012},"timestamp":123}')  # 0.000012

    def testToJsonStringWithFloatValueWithExponent6(self):
        txBody = TransactionBody(
            operation=SetOperation(
                ref="/afan",
                value=0.0000012,  # '1.2e-06'
                type="SET_VALUE",
            ),
            nonce=10,
            timestamp=123,
        )
        self.assertEqual(
            toJsonString(txBody),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":0.0000012},"timestamp":123}')  # 0.0000012

    def testToJsonStringWithFloatValueWithExponent7(self):
        txBody = TransactionBody(
            operation=SetOperation(
                ref="/afan",
                value=0.00000012,  # '1.2e-07'
                type="SET_VALUE",
            ),
            nonce=10,
            timestamp=123,
        )
        self.assertEqual(
            toJsonString(txBody),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":1.2e-07},"timestamp":123}')  # 1.2e-07

    def testToJsonStringWithNegativeFloatValueWithExponent5(self):
        txBody = TransactionBody(
            operation=SetOperation(
                ref="/afan",
                value=-0.000012,  # '-1.2e-05'
                type="SET_VALUE",
            ),
            nonce=10,
            timestamp=123,
        )
        self.assertEqual(
            toJsonString(txBody),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-0.000012},"timestamp":123}')  # -0.000012

    def testToJsonStringWithNegativeFloatValueWithExponent6(self):
        txBody = TransactionBody(
            operation=SetOperation(
                ref="/afan",
                value=-0.0000012,  # '-1.2e-06'
                type="SET_VALUE",
            ),
            nonce=10,
            timestamp=123,
        )
        self.assertEqual(
            toJsonString(txBody),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-0.0000012},"timestamp":123}')  # -0.0000012

    def testToJsonStringWithNegativeFloatValueWithExponent7(self):
        txBody = TransactionBody(
            operation=SetOperation(
                ref="/afan",
                value=-0.00000012,  # '-1.2e-07'
                type="SET_VALUE",
            ),
            nonce=10,
            timestamp=123,
        )
        self.assertEqual(
            toJsonString(txBody),
            '{"nonce":10,"operation":{"ref":"/afan","type":"SET_VALUE","value":-1.2e-07},"timestamp":123}')  # -1.2e-07

class TestByteToHex(TestCase):
    def testByteToHex(self):
        byt = bytes.fromhex("5b9ac8")
        hex = bytesToHex(byt)
        self.assertEqual(hex, "0x5b9ac8")

    def testByteToHexEmptyBytes(self):
        byt = bytes(0)
        hex = bytesToHex(byt)
        self.assertEqual(hex, "0x")

class TestIsValidPrivate(TestCase):
    SECP256K1_N: int = 0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141

    def testIsValidPrivateShortInput(self):
        tmp = "0011223344"
        self.assertFalse(isValidPrivate(bytes.fromhex(tmp)))

    def testIsValidPrivateTooBigInput(self):
        tmp = "3a443d8381a6798a70c6ff9304bdc8cb0163c23211d11628fae52ef9e0dca11a001cf066d56a8156fc201cd5df8a36ef694eecd258903fca7086c1fae7441e1d"
        self.assertFalse(isValidPrivate(bytes.fromhex(tmp)))
    
    def testIsValidPrivateInvalidCurveZero(self):
        tmp = "0000000000000000000000000000000000000000000000000000000000000000"
        self.assertFalse(isValidPrivate(bytes.fromhex(tmp)))

    def testIsValidPrivateInvalidCurveEqN(self):
        self.assertFalse(isValidPrivate(self.SECP256K1_N.to_bytes(32, "big")))
    
    def testIsValidPrivateInvalidCurveGtN(self):
        self.assertFalse(isValidPrivate((self.SECP256K1_N + 1).to_bytes(32, "big")))

    def testIsValidPrivateWork(self):
        self.assertTrue(isValidPrivate((self.SECP256K1_N - 1).to_bytes(32, "big")))

class TestIsValidPublic(TestCase):
    def testIsValid(self):
        pubKey = "3a443d8381a6798a70c6ff9304bdc8cb0163c23211d11628fae52ef9e0dca11a001cf066d56a8156fc201cd5df8a36ef694eecd258903fca7086c1fae7441e1d"
        self.assertTrue(isValidPublic(bytes.fromhex(pubKey)))

    def testIsValidPublicShortInput(self):
        pubKey = "3a443d8381a6798a70c6ff9304bdc8cb0163c23211d11628fae52ef9e0dca11a001cf066d56a8156fc201cd5df8a36ef694eecd258903fca7086c1fae744"
        self.assertFalse(isValidPublic(bytes.fromhex(pubKey)))
    
    def testIsValidPublicBigInput(self):
        pubKey = "3a443d8381a6798a70c6ff9304bdc8cb0163c23211d11628fae52ef9e0dca11a001cf066d56a8156fc201cd5df8a36ef694eecd258903fca7086c1fae7441e1d00"
        self.assertFalse(isValidPublic(bytes.fromhex(pubKey)))

    def testIsValidPublicSEC1Key(self):
        pubKey = "043a443d8381a6798a70c6ff9304bdc8cb0163c23211d11628fae52ef9e0dca11a001cf066d56a8156fc201cd5df8a36ef694eecd258903fca7086c1fae7441e1d"
        self.assertFalse(isValidPublic(bytes.fromhex(pubKey)))
        self.assertTrue(isValidPublic(bytes.fromhex(pubKey), True))

    def testIsValidPublicInvalidSEC1Key(self):
        pubKey = "023a443d8381a6798a70c6ff9304bdc8cb0163c23211d11628fae52ef9e0dca11a001cf066d56a8156fc201cd5df8a36ef694eecd258903fca7086c1fae7441e1d"
        self.assertFalse(isValidPublic(bytes.fromhex(pubKey), True))

    def testIsValidPublicCompressedSEC1Key(self):
        pubKey = "033a443d8381a6798a70c6ff9304bdc8cb0163c23211d11628fae52ef9e0dca11a"
        self.assertTrue(isValidPublic(bytes.fromhex(pubKey), True))

class TestIsValidAddress(TestCase):
    def testIsValidAddress(self):
        self.assertTrue(isValidAddress(address))
        self.assertFalse(isValidAddress(address[1:]))

    def testIsValidAddressNonChecksummed(self):
        self.assertTrue(isValidAddress("0x" + address[2:].upper()))
        self.assertTrue(isValidAddress(address.lower()))

class TestAreSameAddress(TestCase):
    def testAreSameAddress(self):
        self.assertTrue(address, "0x" + address[2:].upper())
        self.assertTrue(address, address.lower())

class TestPubToAddress(TestCase):
    def testPubToAddress(self):
        pubKey = "3a443d8381a6798a70c6ff9304bdc8cb0163c23211d11628fae52ef9e0dca11a001cf066d56a8156fc201cd5df8a36ef694eecd258903fca7086c1fae7441e1d"
        address = "2f015c60e0be116b1f0cd534704db9c92118fb6a"
        self.assertEqual(pubToAddress(bytes.fromhex(pubKey)).hex(), address)

    def testPubToAddressSEC1(self):
        pubKey = "043a443d8381a6798a70c6ff9304bdc8cb0163c23211d11628fae52ef9e0dca11a001cf066d56a8156fc201cd5df8a36ef694eecd258903fca7086c1fae7441e1d"
        address = "2f015c60e0be116b1f0cd534704db9c92118fb6a"
        self.assertEqual(pubToAddress(bytes.fromhex(pubKey), True).hex(), address)

    def testPubToAddressInvalid(self):
        pubKey = "3a443d8381a6798a70c6ff9304bdc8cb0163c23211d11628fae52ef9e0dca11a001cf066d56a8156fc201cd5df8a36ef694eecd258903fca7086c1fae744"
        with self.assertRaises(ValueError):
            pubToAddress(bytes.fromhex(pubKey))
        
    def testPubToAddress0x(self):
        pubKey = "0x3a443d8381a6798a70c6ff9304bdc8cb0163c23211d11628fae52ef9e0dca11a001cf066d56a8156fc201cd5df8a36ef694eecd258903fca7086c1fae7441e1d"
        address = "2f015c60e0be116b1f0cd534704db9c92118fb6a"
        self.assertEqual(pubToAddress(pubKey).hex(), address)

class TestAddHexPrefix(TestCase):
    string = "d658a4b8247c14868f3c512fa5cbb6e458e4a989"
    def testAddHexPrefix(self):
        self.assertEqual(addHexPrefix(self.string), "0x" + self.string)
        with self.assertRaises(AttributeError):
            addHexPrefix(1)

class TestPrivateToPublic(TestCase):
    def testPrivateToPublic(self):
        self.assertEqual(privateToPublic(sk), pk)

class TestPrivateToAddress(TestCase):
    def testPrivateToAddress(self):
        self.assertEqual(privateToAddress(sk), address)

class TestV3KeyStore(TestCase):
    def testV3KeyStore(self):
        v3keystore = V3Keystore.fromPrivateKey(sk, v3KeystorePassword)
        self.assertEqual(v3keystore.toPrivateKey(v3KeystorePassword), sk)
    
    def testV3KeyStoreKDF(self):
        for kdf in v3KeystoreKDFList:
            v3keystore = V3Keystore.fromPrivateKey(sk, v3KeystorePassword, V3KeystoreOptions(kdf=kdf))
            self.assertEqual(v3keystore.toPrivateKey(v3KeystorePassword), sk, msg=json)

    def testV3KeyStoreCipher(self):
        for cipher in v3KeystoreCipherList:
            v3keystore = V3Keystore.fromPrivateKey(sk, v3KeystorePassword, V3KeystoreOptions(cipher=cipher))
            self.assertEqual(v3keystore.toPrivateKey(v3KeystorePassword), sk, msg=json)

    def testV3KeyStoreFromJSON(self):
        for json in v3KeystoreJSONList:
            v3keystoreFromJSON = V3Keystore.fromJSON(json)
            self.assertEqual(v3keystoreFromJSON.toPrivateKey(v3KeystorePassword), sk, msg=json)

    def testV3KeyStoreFromFixed(self):
        v3keystore = V3Keystore.fromPrivateKey(
            sk, v3KeystorePassword,
            V3KeystoreOptions(uuid=v3KeystoreFixedUUID, iv=v3KeystoreFixedIV, salt=v3KeystoreFixedSalt)
        )
        self.assertEqual(str(v3keystore), v3KeystoreFixedJSON)
        
class TestToBytes(TestCase):
    def testToBytes(self):
        # bytes
        self.assertEqual(toBytes(bytes(0)), bytes(0))
        # array
        self.assertEqual(toBytes([]), bytes(0))
        # str
        self.assertEqual(toBytes("11"), bytes([49, 49]))
        self.assertEqual(toBytes("0x11"), bytes([17]))
        self.assertEqual(toBytes("1234").hex(), "31323334")
        self.assertEqual(toBytes("0x1234").hex(), "1234")
        # int
        self.assertEqual(toBytes(1), bytes([1]))
        # None
        self.assertEqual(toBytes(None), bytes(0))
    
    def testToBytesFail(self):
        with self.assertRaises(TypeError):
            toBytes({"test": 1})

class TestHashTransaction(TestCase):
    def testHashTransaction(self):
        txHash = hashTransaction(tx)
        differentHash = hashTransaction(txDifferent)
        self.assertNotEqual(txHash, differentHash)

class TestHashMessage(TestCase):
    def testHashMessage(self):
        hash = hashMessage(message)
        self.assertEqual(hash, bytes.fromhex("14894951ffca216088cba18b434a31fe88cd706886c5f64e0582711d57757ed6"))

class TestToChecksumAddress(TestCase):
    def testToChecksumAddress(self):
        for address in checksumAddresses:
            self.assertEqual(toChecksumAddress(address.lower()), address)

class TestEcSignTransaction(TestCase):
    def testEcSignTransaction(self):
        signature = ecSignTransaction(tx, sk)
        self.assertTrue(ecVerifySig(tx, signature, address))

class TestEcSignMessage(TestCase):
    def testEcSignMessage(self):
        signature = ecSignMessage(message, sk)
        self.assertTrue(ecVerifySig(message, signature, address))

class TestEcVerifySig(TestCase):
    def testEcVerifySig(self):
        self.assertTrue(ecVerifySig(message, correctSignature, address))
    
    def testEcVerifySigCorruptSignature(self):
        corruptSignature = "0x2" + correctSignature[3:]
        self.assertFalse(ecVerifySig(message, corruptSignature, address))
    
    def testEcVerifySigIncorrectAddress(self):
        self.assertFalse(ecVerifySig(message, correctSignature, "0x1D982a9bb0A224618Ac369Ff0a6B8B1c51E5c472"))

    def testEcVerifySigDifferentMessage(self):
        self.assertFalse(ecVerifySig("Hello World", correctSignature, address))

class TestMnemonicToPrivatekey(TestCase):
    def testMnemonicToPrivatekey(self):
        self.assertEqual(mnemonicToPrivatekey(mnemonic), mnemonicPrivateKey)

class TestEncryption(TestCase):
    def testEncryptToDecrypt(self):
        tmpEncrypted = encryptWithPublicKey(pk, message)
        decrypted = decryptWithPrivateKey(sk, tmpEncrypted)
        self.assertEqual(message, decrypted)

    def testDecryptEncrypted(self):
        decrypted = decryptWithPrivateKey(sk, encrypted)
        self.assertEqual(message, decrypted)

# TODO(kriii): Add tests after implement `encode`.
# class TestEncode(TestCase):

# TODO(kriii): Add tests after implement `decode`.
# class TestDecode(TestCase):




from abc import *
from typing import List, Any
from ain.types import TransactionInput, TransactionBody

class Signer(metaclass = ABCMeta):
    """An abstract class for signing messages and transactions.
    """
    
    @abstractmethod
    def getAddress(self, address: str = None) -> str:
        """Gets an account's checksum address.
           If the address is not given, the default account of the signer is used.

        Args:
            address (str, Optional): The address of the account.

        Returns:
            str: The checksum address.
        """
        pass

    @abstractmethod
    async def signMessage(self, message: Any, address: str = None) -> str:
        """Signs a message using an account.
           If an address is not given, the default account of the signer is used.

        Args:
            message (Any): The message to sign.
            address (str, Optional): The address of the account.

        Returns:
            str: The signature.
        """
        pass

    @abstractmethod
    async def sendTransaction(self, transactionObject: TransactionInput, isDryrun = False) -> Any:
        """Signs and sends a transaction to the network.

        Args:
            transactionObject (TransactionInput): The transaction input object.
            isDryrun (bool): The dryrun option.

        Returns:
            The return value of the blockchain API.
        """
        pass

    @abstractmethod
    async def sendTransactionBatch(self, transactionObjects: List[TransactionInput]) -> List[Any]:
        """Signs and sends multiple transactions in a batch to the network.

        Args:
            transactionObjects (List[TransactionInput]): The list of the transaction input objects.
        
        Returns:
            The return value of the blockchain API.
        """
        pass

    @abstractmethod
    async def sendSignedTransaction(self, signature: str, txBody: TransactionBody, isDryrun = False) -> Any:
        """Sends a signed transaction to the network.

        Args:
            signature (str): The signature.
            txBody (TransactionBody): The transaction body.
            isDryrun (bool): The dryrun option.

        Returns:
            The return value of the blockchain API.
        """
        pass



import aiohttp
import json
from urllib.parse import urlparse, urljoin
from jsonrpcclient import request, parse, Ok, Error
from typing import TYPE_CHECKING
from ain.errors import BlockchainError

if TYPE_CHECKING:
    from ain.ain import Ain

JSON_RPC_ENDPOINT = "json-rpc"

class Provider:
    """Class for the AIN Blockchain node provider."""

    endPoint: str
    """The endpoint of the node provider."""
    apiEndPoint: str
    """The JSON-RPC API endpoint of the node provider."""
    _ain: "Ain"
    
    def __init__(self, ain: "Ain", endpoint: str):
        self._ain = ain
        parsed = urlparse(endpoint)
        if parsed.scheme == "" or parsed.netloc == "":
            raise ValueError("Invalid endpoint received.")
        self.endPoint = parsed.geturl()
        self.apiEndPoint = urljoin(self.endPoint, JSON_RPC_ENDPOINT)

    async def send(self, rpcMethod: str, params: dict = {}):
        """Creates the JSON-RPC payload and sends it to the node.

        Args:
            rpcMethod (str): The JSON-RPC method name.
            params (dict): The parameters for the payload.

        Returns:
            The response result.
        """
        params.update({"protoVer": self._ain.net.protoVer})
        dump = json.dumps(
            request(rpcMethod, params=params, id=0),
            default=lambda o: o.__dict__,
            separators=(",", ":"),
            sort_keys=True
        )
        async with aiohttp.ClientSession() as session:
            async with session.post(self.apiEndPoint, json=json.loads(dump)) as response:
                res = await response.read()
                parsed = parse(json.loads(res))
                if isinstance(parsed, Ok):
                    rawResult = parsed.result
                    if self._ain.rawResultMode:
                        return rawResult
                    if not isinstance(rawResult, dict) or not ("code" not in rawResult or rawResult["code"] == 0):
                        raise BlockchainError(rawResult["code"], rawResult["message"])
                    if "code" in rawResult:
                        return rawResult
                    return rawResult.get("result", None)
                if isinstance(parsed, Error):
                    return parsed
                raise AssertionError("The response must be singular.")



from typing import List, Any, Union
from ain.provider import Provider
from ain.net import Network
from ain.wallet import Wallet
from ain.types import AinOptions, TransactionInput, TransactionBody, ValueOnlyTransactionInput
from ain.db import Database
from ain.signer import Signer
from ain.signer.default_signer import DefaultSigner
from ain.utils import getTimestamp, toChecksumAddress

class Ain:
    """Class for the `Ain` instance, the interactor for the AIN Blockchain.

    Args:
        providerUrl (str): The URL of the provider, the listening blockchain node.
        chainId (int): The chain ID of the provider. Defaults to 0.
        ainOptions (AinOptions): The options for this `Ain` instance. Defaults to no options.
    """
    
    provider: Provider
    """The `Provider` instance."""
    chainId: int
    """The chain ID of the provider."""
    rawResultMode: bool
    """The flag of the raw result mode."""
    db: Database
    """The `Database` instance."""
    net: Network
    """The `Network` instance."""
    wallet: Wallet
    """The `Wallet` instance."""
    signer: Signer
    """The `Signer` instance."""

    def __init__(self, providerUrl: str, chainId: int = 0, ainOptions: AinOptions = AinOptions()):
        self.provider = Provider(self, providerUrl)
        self.chainId = chainId
        self.rawResultMode = getattr(ainOptions, "rawResultMode", False)
        self.net = Network(self.provider)
        self.wallet = Wallet(self, self.chainId)
        self.db = Database(self, self.provider)
        self.signer = DefaultSigner(self.wallet, self.provider)

    def setProvider(self, providerUrl: str, chainId: int = 0):
        """Sets a new provider

        Args:
            providerUrl (str): The URL of the new provider.
            chainId (int): The chain ID of the new provider. Defaults to 0.
        """
        self.provider = Provider(self, providerUrl)
        self.chainId = chainId
        self.net = Network(self.provider)
        self.wallet.chainId = chainId
        self.db = Database(self, self.provider)

    def setSigner(self, signer: Signer):
        """Sets a new signer

        Args:
            signer (Signer): The signer to set.
        """
        self.signer = signer

    # TODO(kriii): Return objects typing.

    async def getBlock(
        self,
        blockHashOrBlockNumber: Union[str, int],
        returnTransactionObjects: bool = False,
    ) -> Any:
        """Gets a block with the given hash or block number.

        Args:
            blockHashOrBlockNumber (Union[str, int]): The block hash or the block number.
            returnTransactionObjects (bool): If `True`, returns the full transaction objects.
                If `False`, returns only the transaction hashes. Default to `False`.
        
        Returns:
            The block with the given hash or block number.
        """
        if type(blockHashOrBlockNumber) is str:
            return await self.provider.send(
                "ain_getBlockByHash",
                {
                    "getFullTransactions": returnTransactionObjects,
                    "hash": blockHashOrBlockNumber,
                },
            )
        elif type(blockHashOrBlockNumber) is int:
            return await self.provider.send(
                "ain_getBlockByNumber",
                {
                    "getFullTransactions": returnTransactionObjects,
                    "number": blockHashOrBlockNumber,
                },
            )
        else:
            raise TypeError("blockHashOrBlockNumber has invalid type")

    async def getProposer(
        self,
        blockHashOrBlockNumber: Union[str, int],
    ) -> Any:
        """Gets an address of the proposer of the given block.

        Args:
            blockHashOrBlockNumber (Union[str, int]): The block hash or the block number.

        Returns:
            The address of the proposer of the given block.
        """
        if type(blockHashOrBlockNumber) is str:
            return await self.provider.send(
                "ain_getProposerByHash", {"hash": blockHashOrBlockNumber}
            )
        elif type(blockHashOrBlockNumber) is int:
            return await self.provider.send(
                "ain_getProposerByNumber", {"number": blockHashOrBlockNumber}
            )
        else:
            raise TypeError("blockHashOrBlockNumber has invalid type")

    async def getValidators(
        self,
        blockHashOrBlockNumber: Union[str, int],
    ) -> Any:
        """Gets the list of validators for a given block

        Args:
            blockHashOrBlockNumber (Union[str, int]): The block hash or the block number.
            
        Returns:
            The list of validators for a given block.
        """
        if type(blockHashOrBlockNumber) is str:
            return await self.provider.send(
                "ain_getValidatorsByHash", {"hash": blockHashOrBlockNumber}
            )
        elif type(blockHashOrBlockNumber) is int:
            return await self.provider.send(
                "ain_getValidatorsByNumber", {"number": blockHashOrBlockNumber}
            )
        else:
            raise TypeError("blockHashOrBlockNumber has invalid type")

    async def getTransaction(self, transactionHash: str) -> Any:
        """Gets a transaction with the given transaction hash.

        Args:
            transactionHash (str): The transaction hash.

        Returns:
            The transaction with the given transaction hash.
        """
        return await self.provider.send("ain_getTransactionByHash", {"hash": transactionHash})

    async def getStateUsage(self, appName: str) -> Any:
        """Gets a state usage with the given app name.

        Args:
            appName (str): The app name;

        Returns:
            The state usage with the given app name.
        """
        return await self.provider.send("ain_getStateUsage", {"app_name": appName})

    async def validateAppName(self, appName: str) -> Any:
        """Validate a given app name.

        Args:
            appName (str): The app name;

        Returns:
            The validity of the given app name.
        """
        return await self.provider.send("ain_validateAppName", {"app_name": appName})

    async def sendTransaction(self, transactionObject: TransactionInput, isDryrun = False) -> Any:
        """Signs and sends a transaction to the network.

        Args:
            transactionObject (TransactionInput): The transaction input object.
            isDryrun (bool): The dryrun option.

        Returns:
            The return value of the blockchain API.
        """
        return await self.signer.sendTransaction(transactionObject, isDryrun)

    async def sendSignedTransaction(self, signature: str, txBody: TransactionBody, isDryrun = False) -> Any:
        """Sends a signed transaction to the network.

        Args:
            signature (str): The signature.
            txBody (TransactionBody): The transaction body.
            isDryrun (bool): The dryrun option.

        Returns:
            The return value of the blockchain API.
        """
        return await self.signer.sendSignedTransaction(signature, txBody, isDryrun)

    async def sendTransactionBatch(self, transactionObjects: List[TransactionInput]) -> List[Any]:
        """Signs and sends multiple transactions in a batch to the network.

        Args:
            transactionObjects (List[TransactionInput]): The list of the transaction input objects.
        
        Returns:
            The return value of the blockchain API.
        """
        return await self.signer.sendTransactionBatch(transactionObjects)
        
    def depositConsensusStake(self, input: ValueOnlyTransactionInput) -> Any:
        """Sends a transaction that deposits AIN for consensus staking.

        Args:
            input (ValueOnlyTransactionInput): The value only transaction input to deposit AIN value.
            
        Returns:
            The transaction results.
        """
        return self.__stakeFunction("/deposit/consensus", input)

    def withdrawConsensusStake(self, input: ValueOnlyTransactionInput) -> Any:
        """
        Sends a transaction that withdraws AIN for consensus staking.

        Args:
            input (ValueOnlyTransactionInput): The value only transaction input to withdraw AIN value.
            
        Returns:
            The transaction results.
        """
        return self.__stakeFunction("/withdraw/consensus", input)

    async def getConsensusStakeAmount(self, account: str = None) -> Any:
        """Gets an amount of the AIN currently staked for participating in consensus protocol.

        Args:
            Account (str, Optional): The address of the consensus participant.

        Returns:
            The amount of the AIN of that address.
        """
        if account is None:
            address = self.signer.getAddress()
        else:
            address = toChecksumAddress(account)
        return await self.db.ref(f"/deposit_accounts/consensus/{address}").getValue()

    def __stakeFunction(self, path: str, input: ValueOnlyTransactionInput) -> Any:
        """A base function for all staking related database changes. It builds a
        deposit/withdraw transaction and sends the transaction by calling sendTransaction().
        """
        if not hasattr(input, "value"):
            raise ValueError("a value should be specified.")
        if type(input.value) is not int:
            raise ValueError("value has to be a int.")
        input.address = self.signer.getAddress(getattr(input, "address", None))
        ref = self.db.ref(f'{path}/{input.address}').push()
        input.ref = "value"
        return ref.setValue(input)
